/*
 * Copyright 2010 Alibaba Group Holding Limited.
 * All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.alibaba.citrus.service.requestcontext.rundata.impl;

import static com.alibaba.citrus.service.requestcontext.util.RequestContextUtil.*;
import static com.alibaba.citrus.util.Assert.*;

import java.io.IOException;

import javax.servlet.http.HttpSession;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.alibaba.citrus.service.requestcontext.RequestContext;
import com.alibaba.citrus.service.requestcontext.buffered.BufferedRequestContext;
import com.alibaba.citrus.service.requestcontext.lazycommit.LazyCommitRequestContext;
import com.alibaba.citrus.service.requestcontext.locale.SetLocaleRequestContext;
import com.alibaba.citrus.service.requestcontext.parser.CookieParser;
import com.alibaba.citrus.service.requestcontext.parser.ParameterParser;
import com.alibaba.citrus.service.requestcontext.parser.ParserRequestContext;
import com.alibaba.citrus.service.requestcontext.rundata.RunData;
import com.alibaba.citrus.service.requestcontext.rundata.User;
import com.alibaba.citrus.service.requestcontext.support.AbstractRequestContextWrapper;
import com.alibaba.citrus.util.StringUtil;
import com.alibaba.citrus.util.io.ByteArray;

/**
 * <code>RunData</code>ʵࡣ
 * 
 * @author Michael Zhou
 */
public class RunDataImpl extends AbstractRequestContextWrapper implements RunData {
    private final static Logger log = LoggerFactory.getLogger(RunData.class);
    private final BufferedRequestContext bufferedRequestContext;
    private final LazyCommitRequestContext lazyCommitRequestContext;
    private final SetLocaleRequestContext setLocaleRequestContext;
    private final ParserRequestContext parserRequestContext;
    private User user;

    public RunDataImpl(RequestContext wrappedContext) {
        super(wrappedContext);
        this.bufferedRequestContext = findRequestContext(wrappedContext, BufferedRequestContext.class);
        this.lazyCommitRequestContext = findRequestContext(wrappedContext, LazyCommitRequestContext.class);
        this.setLocaleRequestContext = findRequestContext(wrappedContext, SetLocaleRequestContext.class);
        this.parserRequestContext = findRequestContext(wrappedContext, ParserRequestContext.class);

        if (bufferedRequestContext == null) {
            log.debug("RunData feature BufferedRequestContext disabled");
        }

        if (lazyCommitRequestContext == null) {
            log.debug("RunData feature LazyCommitRequestContext disabled");
        }

        if (setLocaleRequestContext == null) {
            log.debug("RunData feature SetLocaleRequestContext disabled");
        }

        if (parserRequestContext == null) {
            log.debug("RunData feature ParserRequestContext disabled");
        }
    }

    protected BufferedRequestContext getBufferedRequestContext() {
        return assertNotNull(bufferedRequestContext, "Could not find BufferedRequestContext in request context chain");
    }

    protected LazyCommitRequestContext getLazyCommitRequestContext() {
        return assertNotNull(lazyCommitRequestContext,
                "Could not find LazyCommitRequestContext in request context chain");
    }

    protected SetLocaleRequestContext getSetLocaleRequestContext() {
        return assertNotNull(setLocaleRequestContext, "Could not find SetLocaleRequestContext in request context chain");
    }

    protected ParserRequestContext getParserRequestContext() {
        return assertNotNull(parserRequestContext, "Could not find ParserRequestContext in request context chain");
    }

    // ===================================================
    // HTTP requestϢ
    // ===================================================

    /**
     * ȡqueryһִд˷ʱrequestȡеĲ
     * 
     * @return <code>ParameterParser</code>ʵ
     */
    public ParameterParser getParameters() {
        return getParserRequestContext().getParameters();
    }

    /**
     * ȡcookieһִд˷ʱrequestȡcookies
     * 
     * @return <code>CookieParser</code>ʵ
     */
    public CookieParser getCookies() {
        return getParserRequestContext().getCookies();
    }

    /**
     * ȡõǰHTTP session
     * 
     * @return HTTP session
     */
    public HttpSession getSession() {
        return getRequest().getSession();
    }

    /**
     * ȡwebӦõ·൱<code>HttpServletRequest.getContextPath</code>صֵ
     * 
     * @return webӦõ·
     */
    public String getContextPath() {
        return getRequest().getContextPath();
    }

    /**
     * ȡservlet·൱<code>HttpServletRequest.getServletPath</code>صֵ
     * 
     * @return servlet·
     */
    public String getServletPath() {
        return getRequest().getServletPath();
    }

    /**
     * ȡpath info·൱<code>HttpServletRequest.getPathInfo</code>صֵ
     * 
     * @return path info·
     */
    public String getPathInfo() {
        return getRequest().getPathInfo();
    }

    /**
     * ȡõǰrequest URLquery string
     * 
     * @return ǰrequest URL
     */
    public String getRequestURL() {
        return getRequestURL(true);
    }

    /**
     * ȡõǰrequest URLquery string
     * 
     * @param withQueryString Ƿquery string
     * @return ǰrequest URL
     */
    public String getRequestURL(boolean withQueryString) {
        StringBuffer buffer = getRequest().getRequestURL();

        if (withQueryString) {
            String queryString = StringUtil.trimToNull(getRequest().getQueryString());

            if (queryString != null) {
                buffer.append('?').append(queryString);
            }
        }

        return buffer.toString();
    }

    /**
     * ǷΪpost
     */
    public boolean isPostRequest() {
        return "post".equalsIgnoreCase(getRequest().getMethod());
    }

    // ===================================================
    // Ϣ
    // ===================================================

    /**
     * ȡڷʵǰӦõû
     * 
     * @return û
     */
    public User getUser() {
        return this.user;
    }

    /**
     * ڷʵǰӦõû
     * 
     * @param user û
     */
    public void setUser(User user) {
        this.user = user;
    }

    /**
     * ȡú͵ǰ󶨵Ķ󡣵ʱеattributes
     * 
     * @param key key
     * @return keyӦĶ
     */
    public Object getAttribute(String key) {
        return getRequest().getAttribute(key);
    }

    /**
     * ָ󶨵ǰСʱеattributes
     * 
     * @param key key
     * @param object keyӦĶ
     */
    public void setAttribute(String key, Object object) {
        if (object == null) {
            getRequest().removeAttribute(key);
        } else {
            getRequest().setAttribute(key, object);
        }
    }

    // ===================================================
    // HTTP responseϢ
    // ===================================================

    /**
     * ȡcontent type
     * 
     * @return content typecharsetĶ
     */
    public String getContentType() {
        return getSetLocaleRequestContext().getResponseContentType();
    }

    /**
     * content type content typecharset
     * <code>getCharacterEncoding</code>ãcharsetǡ
     * 
     * @param contentType content type
     */
    public void setContentType(String contentType) {
        getResponse().setContentType(contentType);
    }

    /**
     * content type content typecharset
     * <code>getCharacterEncoding</code>ãcharsetǡ
     * <p>
     * <code>appendCharset</code>Ϊ<code>false</code>content
     * typeнcharsetǡ
     * </p>
     * 
     * @param contentType content type
     * @param appendCharset ַ
     */
    public void setContentType(String contentType, boolean appendCharset) {
        getSetLocaleRequestContext().setResponseContentType(contentType, appendCharset);
    }

    /**
     * ȡresponseַ
     */
    public String getCharacterEncoding() {
        return getResponse().getCharacterEncoding();
    }

    /**
     * responseַע⣬˷ڵһ<code>getWriter</code>֮ǰִС
     * 
     * @param charset ַcharsetΪ<code>null</code>
     *            contentTypeɾcharset
     */
    public void setCharacterEncoding(String charset) {
        getSetLocaleRequestContext().setResponseCharacterEncoding(charset);
    }

    /**
     * ȡضURI
     * 
     * @return ضURIûض򷵻<code>null</code>
     */
    public String getRedirectLocation() {
        return getLazyCommitRequestContext().getRedirectLocation();
    }

    /**
     * ضURI
     * 
     * @param location ضURI
     * @throws IOException ʧ
     * @throws IllegalStateException responseѾcommitted
     */
    public void setRedirectLocation(String location) throws IOException {
        getResponse().sendRedirect(location);
    }

    /**
     * оϵͳǷѾض
     * 
     * @return <code>setRedirectLocation</code>ã򷵻<code>true</code>
     */
    public boolean isRedirected() {
        return getLazyCommitRequestContext().isRedirected();
    }

    /**
     * ȡõHTTP status
     * 
     * @return HTTP statusֵ
     */
    public int getStatusCode() {
        return getLazyCommitRequestContext().getStatus();
    }

    /**
     * HTTP status
     * 
     * @param status HTTP statusֵ
     */
    public void setStatusCode(int status) {
        getResponse().setStatus(status);
    }

    // ===================================================
    // Response bufferơ
    // ===================================================

    /**
     * ǷϢڴС
     * 
     * @return ǣ򷵻<code>true</code>
     */
    public boolean isBuffering() {
        return getBufferedRequestContext().isBuffering();
    }

    /**
     * bufferģʽó<code>true</code>ʾϢڴУֱԭʼresponseС
     * <p>
     * ˷<code>getOutputStream</code><code>getWriter</code>֮ǰִУ׳
     * <code>IllegalStateException</code>
     * </p>
     * 
     * @param buffering Ƿbuffer
     * @throws IllegalStateException <code>getOutputStream</code>
     *             <code>getWriter</code>Ѿִ
     */
    public void setBuffering(boolean buffering) {
        getBufferedRequestContext().setBuffering(buffering);
    }

    /**
     * µbufferϵbuffer
     * 
     * @throws IllegalStateException bufferģʽ<code>getWriter</code>
     *             <code>getOutputStream</code>δ
     */
    public void pushBuffer() {
        getBufferedRequestContext().pushBuffer();
    }

    /**
     * bufferջֻһbuffer򵯳ٴһµġ
     * 
     * @return buffer
     * @throws IllegalStateException bufferģʽ<code>getWriter</code>
     *             ã<code>getOutputStream</code>δ
     */
    public ByteArray popByteBuffer() {
        return getBufferedRequestContext().popByteBuffer();
    }

    /**
     * bufferջֻһbuffer򵯳ٴһµġ
     * 
     * @return buffer
     * @throws IllegalStateException bufferģʽ<code>getOutputStream</code>
     *             ã<code>getWriter</code>δ
     */
    public String popCharBuffer() {
        return getBufferedRequestContext().popCharBuffer();
    }

    /**
     * buffersʾϢ
     * 
     * @throws IllegalStateException responseѾcommit
     */
    public void resetBuffer() {
        getResponse().resetBuffer();
    }

    /**
     * ַָ<code>getCaseFolding()</code>ãתָСдʽ
     * 
     * @param str Ҫתַ
     * @return תַ
     */
    public String convertCase(String str) {
        return getParserRequestContext().convertCase(str);
    }
}
